// Aplica o back-end do juninho
@target          : mips 
// Especifica o tipo de metodo empregado para salvar o resultado da compilação
@export          : simulation //MultCoreJun

// Diretorio onde o resultado da compilação será gravado
@outputDirectory : `C:\Users\EUGENIO CARVALHO\Desktop\tmp\laplaciano`

// Quantidade de palavras de um bloco 32 palavras de 4bytes
@cacheBlockSize: `32`

/**
    Profile de compilacao do back-end
    O formato da string é um json que descreve um array de definições de cada core
    {
        "stackBaseAddress"   : 5888,     -> endereço da base da pilha do core em questão
        "initFunction": "multiplica(0)", -> define a main de cada core. aceita ate 4 parametros com valores de constantes
        "id": "core0"                    -> label que define o codigo de cada core
    },

    Se "stackBaseAddress" não for definido o compilador irá inferir o valor iniciando do ultimo endereço da memoria    
*/

@profile: `[
    {
        "id"               : "core0",
        "initFunction"     : "multiplica(0)",
        "filename"         : "%s_core_0.txt"
    },
    {
        "id"               : "core1",
        "initFunction"     : "multiplica(5)",
        "filename"         : "%s_core_1.txt"
    },
    {
        "id"               : "core2",
        "initFunction"     : "multiplica(10)",
        "filename"         : "%s_core_2.txt"
    },
    {
        "id"               : "core3",
        "initFunction"     : "multiplica(15)",
        "filename"         : "%s_core_3.txt"
    }
]`


package main;

const (
    LIN = 4
    COL = 4
)

var (
    origin [LIN][COL]int32
    result [LIN][COL]int32
)

func preenche_matriz() {
    var (
        w = 1
        value, LM1, CM1 int32
    )
    
    LM1 = LIN - 1
    CM1 = COL - 1

    for i := 0; i < LIN; i++ {
        for j := 0; j < COL; j++ {
            switch {
                case i == 0, i == LM1: // Primeira  e ultima linha
                    value = 0
                case j == 0, j == CM1: // Primeira e ultima coluna
                    value = 0
                default:               // Elementos do interior da matriz
                    value = w
                    w++
            }
            origin[i][j] = value
        }
    }
}

func laplaciano() {
    var (
        l int32
    )
    LM1 := LIN - 1
    CM1 := COL - 1

    for i:= 1; i < LM1; i++ {
        for j := 1; j < CM1; j++ {
            l = origin[i-1][j]
            l += origin[i][j-1] 
            l += origin[i][j+1] 
            l += origin[i+1][j] 
            l -= (4 * origin[i][j])

            if l > 255 { l = 255 }
            else if l < 0 { l = 0 }

            result[i][j] = l
        }
    }
}

func main() {
    preenche_matriz()
    //laplaciano()
}

/**
Reset por (false) or era label
$$$$$$$$$$$$$$$$ RESET 
Reset por (false) or era label
$$$$$$$$$$$$$$$$ RESET 
SET VAR (_VW1)_S2
$$$$$$$$$$$$$$$$ Get Reverse of @_S2 | {@_S2=_VW1, _VW1=_S2}
Store:_S2:1
$$$$$$$$$$$$$$$$ SetLoaded:_S2:false
SET VAR (_VLM14)_S5
$$$$$$$$$$$$$$$$ Get Reverse of @_S5 | {@_S2=_VW1, _VLM14=_S5, _VW1=_S2, @_S5=_VLM14}
Store:_S5:_T3
$$$$$$$$$$$$$$$$ SetLoaded:_S5:false
SET VAR (_VCM17)_S8
$$$$$$$$$$$$$$$$ Get Reverse of @_S8 | {@_S2=_VW1, _VLM14=_S5, _VCM17=_S8, @_S8=_VCM17, _VW1=_S2, @_S5=_VLM14}
Store:_S8:_T6
$$$$$$$$$$$$$$$$ SetLoaded:_S8:false
SET VAR (_VI9)_S10
$$$$$$$$$$$$$$$$ Get Reverse of @_S10 | {@_S2=_VW1, _VLM14=_S5, _VCM17=_S8, @_S10=_VI9, @_S8=_VCM17, _VI9=_S10, _VW1=_S2, @_S5=_VLM14}
Store:_S10:0
$$$$$$$$$$$$$$$$ SetLoaded:_S10:false
Reset por (true) or era label
$$$$$$$$$$$$$$$$ RESET 
SET VAR (_VJ11)_S12
$$$$$$$$$$$$$$$$ Get Reverse of @_S12 | {_VJ11=_S12, @_S12=_VJ11}
Store:_S12:0
$$$$$$$$$$$$$$$$ SetLoaded:_S12:false
Reset por (true) or era label
$$$$$$$$$$$$$$$$ RESET 
SET VAR (_VI9)_S13
$$$$$$$$$$$$$$$$ LOADED _S13 : false
$$$$$$$$$$$$$$$$ SetLoaded:_S13:true
$$$$$$$$$$$$$$$$ Get Reverse of @_S13 | {@_S13=_VI9, _VI9=_S13}
Reset por (true) or era label
$$$$$$$$$$$$$$$$ RESET 
Reset por (false) or era label
$$$$$$$$$$$$$$$$ RESET 
SET VAR (_VI9)_S14
SET VAR (_VLM14)_S15
$$$$$$$$$$$$$$$$ LOADED _S14 : false
$$$$$$$$$$$$$$$$ SetLoaded:_S14:true
$$$$$$$$$$$$$$$$ Get Reverse of @_S14 | {_VLM14=_S15, @_S14=_VI9, @_S15=_VLM14, _VI9=_S14}
Reset por (true) or era label
$$$$$$$$$$$$$$$$ RESET 
$$$$$$$$$$$$$$$$ LOADED _S15 : false
$$$$$$$$$$$$$$$$ SetLoaded:_S15:true
$$$$$$$$$$$$$$$$ Get Reverse of @_S15 | {}
Reset por (false) or era label
$$$$$$$$$$$$$$$$ RESET 
SET VAR (_VVALUE16)_S17
$$$$$$$$$$$$$$$$ Get Reverse of @_S17 | {@_S17=_VVALUE16, _VVALUE16=_S17}
Store:_S17:0
$$$$$$$$$$$$$$$$ SetLoaded:_S17:false
Reset por (true) or era label
$$$$$$$$$$$$$$$$ RESET 
SET VAR (_VJ11)_S18
$$$$$$$$$$$$$$$$ LOADED _S18 : false
$$$$$$$$$$$$$$$$ SetLoaded:_S18:true
$$$$$$$$$$$$$$$$ Get Reverse of @_S18 | {_VJ11=_S18, @_S18=_VJ11}
Reset por (true) or era label
$$$$$$$$$$$$$$$$ RESET 
Reset por (false) or era label
$$$$$$$$$$$$$$$$ RESET 
SET VAR (_VJ11)_S19
SET VAR (_VCM17)_S20
$$$$$$$$$$$$$$$$ LOADED _S19 : false
$$$$$$$$$$$$$$$$ SetLoaded:_S19:true
$$$$$$$$$$$$$$$$ Get Reverse of @_S19 | {_VCM17=_S20, @_S20=_VCM17, _VJ11=_S19, @_S19=_VJ11}
Reset por (true) or era label
$$$$$$$$$$$$$$$$ RESET 
$$$$$$$$$$$$$$$$ LOADED _S20 : false
$$$$$$$$$$$$$$$$ SetLoaded:_S20:true
$$$$$$$$$$$$$$$$ Get Reverse of @_S20 | {}
Reset por (false) or era label
$$$$$$$$$$$$$$$$ RESET 
SET VAR (_VVALUE16)_S21
$$$$$$$$$$$$$$$$ Get Reverse of @_S21 | {@_S21=_VVALUE16, _VVALUE16=_S21}
Store:_S21:0
$$$$$$$$$$$$$$$$ SetLoaded:_S21:false
Reset por (true) or era label
$$$$$$$$$$$$$$$$ RESET 
SET VAR (_VW1)_S22
SET VAR (_VVALUE16)_S23
$$$$$$$$$$$$$$$$ Get Reverse of @_S23 | {@_S22=_VW1, @_S23=_VVALUE16, _VVALUE16=_S23, _VW1=_S22}
$$$$$$$$$$$$$$$$ LOADED _S22 : false
$$$$$$$$$$$$$$$$ SetLoaded:_S22:true
$$$$$$$$$$$$$$$$ Get Reverse of @_S22 | {@_S22=_VW1, @_S23=_VVALUE16, _VVALUE16=_S23, _VW1=_S22}
Store:_S23:_S22
$$$$$$$$$$$$$$$$ SetLoaded:_S23:false
$$$$$$$$$$$$$$$$ LOADED _S22 : true
$$$$$$$$$$$$$$$$ Get Reverse of @_S22 | {@_S22=_VW1, @_S23=_VVALUE16, _VVALUE16=_S23, _VW1=_S22}
$$$$$$$$$$$$$$$$ LOADED _S22 : true
Reset por (false) or era label
$$$$$$$$$$$$$$$$ RESET 
SET VAR (_VVALUE16)_S24
SET VAR (_VI9)_S25
$$$$$$$$$$$$$$$$ LOADED _S25 : false
$$$$$$$$$$$$$$$$ SetLoaded:_S25:true
$$$$$$$$$$$$$$$$ Get Reverse of @_S25 | {@_S24=_VVALUE16, @_S25=_VI9, _VI9=_S25, _VVALUE16=_S24}
$$$$$$$$$$$$$$$$ LOADED _S25 : true
SET VAR (_VJ11)_S27
$$$$$$$$$$$$$$$$ LOADED _S27 : false
$$$$$$$$$$$$$$$$ SetLoaded:_S27:true
$$$$$$$$$$$$$$$$ Get Reverse of @_S27 | {_VJ11=_S27, @_S24=_VVALUE16, @_S25=_VI9, _VI9=_S25, @_S27=_VJ11, _VVALUE16=_S24}
$$$$$$$$$$$$$$$$ LOADED _S27 : true
SET VAR (_G30)_S31
$$$$$$$$$$$$$$$$ Get Reverse of @_S31 | {@_S31=_G30, _VJ11=_S27, @_S24=_VVALUE16, @_S25=_VI9, _G30=_S31, _VI9=_S25, @_S27=_VJ11, _VVALUE16=_S24}
$$$$$$$$$$$$$$$$ LOADED _S24 : false
$$$$$$$$$$$$$$$$ SetLoaded:_S24:true
$$$$$$$$$$$$$$$$ Get Reverse of @_S24 | {@_S31=_G30, _VJ11=_S27, @_S24=_VVALUE16, @_S25=_VI9, _G30=_S31, _VI9=_S25, @_S27=_VJ11, _VVALUE16=_S24}
Store:_S31[_T29]:_S24
$$$$$$$$$$$$$$$$ SetLoaded:_S31[_T29]:false
$$$$$$$$$$$$$$$$ LOADED _S27 : true
$$$$$$$$$$$$$$$$ Get Reverse of @_S27 | {@_S31=_G30, _VJ11=_S27, @_S24=_VVALUE16, @_S25=_VI9, _G30=_S31, _VI9=_S25, @_S27=_VJ11, _VVALUE16=_S24}
$$$$$$$$$$$$$$$$ LOADED _S27 : true
Reset por (false) or era label
$$$$$$$$$$$$$$$$ RESET 
SET VAR (_VJ11)_S32
$$$$$$$$$$$$$$$$ LOADED _S32 : false
$$$$$$$$$$$$$$$$ SetLoaded:_S32:true
$$$$$$$$$$$$$$$$ Get Reverse of @_S32 | {_VJ11=_S32, @_S32=_VJ11}
Reset por (true) or era label
$$$$$$$$$$$$$$$$ RESET 
SET VAR (_VI9)_S33
$$$$$$$$$$$$$$$$ LOADED _S33 : false
$$$$$$$$$$$$$$$$ SetLoaded:_S33:true
$$$$$$$$$$$$$$$$ Get Reverse of @_S33 | {@_S33=_VI9, _VI9=_S33}
$$$$$$$$$$$$$$$$ Get Reverse of @_S33 | {@_S33=_VI9, _VI9=_S33}
$$$$$$$$$$$$$$$$ LOADED _S33 : true
Reset por (false) or era label
$$$$$$$$$$$$$$$$ RESET 
SET VAR (_VI9)_S34
$$$$$$$$$$$$$$$$ LOADED _S34 : false
$$$$$$$$$$$$$$$$ SetLoaded:_S34:true
$$$$$$$$$$$$$$$$ Get Reverse of @_S34 | {@_S34=_VI9, _VI9=_S34}
Reset por (true) or era label
$$$$$$$$$$$$$$$$ RESET 
Reset por (false) or era label
$$$$$$$$$$$$$$$$ RESET 
*/